/*  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.

 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.zip.DataFormatException;
import java.util.zip.Inflater;


public class Util {
	public static String readString(byte[] input, int offset) {
		int end;
		for(end = offset; end < input.length && input[end] != 0; end++);
		if(end == input.length) throw new RuntimeException("ran out of data");
		String ret = new String(input, offset, end - offset);
		return ret;
	}
	public static byte[] readZlib(RandomAccessFile in, int csize) throws IOException, DataFormatException {
		int len = Util.read32(in);
		byte[] input = new byte[csize - 4];
		in.read(input);
		if(len == -1) return input;
		Inflater i = new Inflater();
		i.setInput(input);
		byte[] ret = new byte[len];
		i.inflate(ret);
		i.reset();
		return ret;
	}
	public static void printBytes(byte[] input, int offset, int length) {
		int read;
		
		StringBuilder b = new StringBuilder();
		for(read = 0; read < length; read++) {
			if(read > 0) {
				if(read % 16 == 0) b.append('\n');
				else b.append(' ');
			}
			int next = input[read+offset] & 0x000000FF;
			if(next < 16) {
				b.append('0');
			}
			b.append(Integer.toHexString(next));
		}
		System.out.println(b.toString());
	}

	public static int read32(byte[] input, int offset) {
		return ((input[offset+0] & 0xFF) << 0) | ((input[offset+1] & 0xFF) << 8) | ((input[offset+2] & 0xFF) << 16) | ((input[offset+3] & 0xFF) << 24);
	}
	public static void write32(byte[] input, int offset, int data) {
		input[offset+0] = (byte) ((data >> 0)&0xFF);
		input[offset+1] = (byte) ((data >> 8)&0xFF);
		input[offset+2] = (byte) ((data >> 16)&0xFF);
		input[offset+3] = (byte) ((data >> 24)&0xFF);
	}
	public static short read16(byte[] input, int offset) {
		return (short) (((input[offset+0] & 0xFF) << 0) | ((input[offset+1] & 0xFF) << 8));
	}
	public static void write32(RandomAccessFile f, int data) throws IOException {
		f.writeByte((data >> 0) & 0xFF);
		f.writeByte((data >> 8) & 0xFF);
		f.writeByte((data >> 16) & 0xFF);
		f.writeByte((data >> 24) & 0xFF);
	}
	public static void write16(RandomAccessFile f, short data) throws IOException {
		f.writeByte((data >> 0) & 0xFF);
		f.writeByte((data >> 8) & 0xFF);
	}
	public static short read16(RandomAccessFile f) throws IOException {
		return (short) (((f.readByte() & 0xFF) << 0) | ((f.readByte() & 0xFF) << 8));
	}
	public static int read32(RandomAccessFile f) throws IOException {
		return ((f.readByte() & 0xFF) << 0) | ((f.readByte() & 0xFF) << 8) | ((f.readByte() & 0xFF) << 16) | ((f.readByte() & 0xFF) << 24);
	}
}
