/*  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.

 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.zip.Deflater;

public class WadMaker {
	private static class FileEntry {
		String filename;
		short id;
		int unknown;
		int offset;
		int csize;
		byte[] data;
	}
	private static class DirectoryEntry {
		final String filename;
		int directoryCount;
		int directoryCountTotal;
		int fileCount;
		int fileCountTotal;
		List<DirectoryEntry> directories = new LinkedList<DirectoryEntry>();
		List<FileEntry> files = new LinkedList<FileEntry>();
		DirectoryEntry(String filename) {
			this.filename = filename;
		}
	}
	
	private static Deflater d = new Deflater(Deflater.BEST_SPEED, false);
	public static byte[] getBytesFromFile(File f) throws IOException {
		FileInputStream i = null;
		try {
			i = new FileInputStream(f);
			byte[] ret = new byte[(int) f.length()];
			i.read(ret);
			return ret;
		} finally {
			if(i != null) i.close();
		}
	}
	private static void traverse(File dirFile, DirectoryEntry dir) throws IOException {
		File[] children = dirFile.listFiles();
		for(File f: children) {
			if(f.isFile()) {
				FileEntry file = new FileEntry();
				file.filename = f.getName().toUpperCase();
				byte[] ufile = getBytesFromFile(f);
				byte[] cfile1 = new byte[ufile.length + 400];
				byte[] cfile2;
				d.reset();
				d.setInput(ufile);
				d.finish();
				int csize = d.deflate(cfile1);
				if(!d.finished()) {
					System.out.println("WARNING: something went wrong?!");
				}
				if(false && csize >= ufile.length) {
					cfile2 = new byte[4 + ufile.length];
					Util.write32(cfile2, 0, -1);
					System.arraycopy(ufile, 0, cfile2, 4, ufile.length);
					file.csize = ufile.length + 4;
					file.data = cfile2;
				} else {
					cfile2 = new byte[4 + csize];
					Util.write32(cfile2, 0, ufile.length);
					System.arraycopy(cfile1, 0, cfile2, 4, csize);
					file.csize = csize + 4;
					file.data = cfile2;		
				}
				dir.fileCount++;
				dir.fileCountTotal++;
				dir.files.add(file);
			}
			if(f.isDirectory()) {
				DirectoryEntry next = new DirectoryEntry(f.getName().toUpperCase());
				dir.directoryCount++;
				dir.directories.add(next);
				traverse(f, next);
				dir.directoryCountTotal++;
				dir.directoryCountTotal += next.directoryCountTotal;
				dir.fileCountTotal += next.fileCountTotal;
			}
		}
	}
	private static int writeFilenamesChild(DirectoryEntry current, Map<String,Integer> filenameMap, int offset, RandomAccessFile out) throws IOException {
		if(!filenameMap.containsKey(current.filename)) {
			filenameMap.put(current.filename, offset);
			byte[] toWrite = current.filename.getBytes();
			out.write(toWrite);
			out.writeByte(0);
			offset += toWrite.length + 1;
		}
		for(FileEntry f: current.files) {
			if(!filenameMap.containsKey(f.filename)) {
				filenameMap.put(f.filename, offset);
				byte[] toWrite = f.filename.getBytes();
				out.write(toWrite);
				out.writeByte(0);
				offset += toWrite.length + 1;
			}
		}
		for(DirectoryEntry d: current.directories) {
			offset = writeFilenamesChild(d, filenameMap, offset, out);
		}
		return offset;
	}
	private static Map<String,Integer> writeFilenames(DirectoryEntry root, RandomAccessFile out) throws IOException {
		long start = out.getFilePointer();
		out.writeInt(0); //we will have to overwrite this
		Map<String,Integer> filenameMap = new HashMap<String,Integer>();
		int total = writeFilenamesChild(root, filenameMap, 0, out);
		while(total % 4 > 0) {
			out.writeByte(0);
			total++;
		}
		out.seek(start);
		Util.write32(out, total);
		out.seek(start + 4 + total);
		return filenameMap;
	}
	private static short traverseFiles1(DirectoryEntry root, RandomAccessFile out, short id) throws IOException {
		for(DirectoryEntry e: root.directories) {
			id = traverseFiles1(e, out, id);
		}
		for(FileEntry e: root.files) {
			e.id = id++;
			Util.write16(out, e.id);
			out.writeByte(0);
			out.writeByte(1);
			Util.write32(out, e.unknown);
		}
		return id;
	}
	private static void traverseWrite(DirectoryEntry root, Map<String,Integer> filenameMap, RandomAccessFile out) throws IOException {
		Util.write32(out, filenameMap.get(root.filename));
		Util.write32(out, root.fileCount);
		Util.write16(out, (short) root.directoryCount);
		out.writeByte(0);
		out.writeByte(0);
		Util.write32(out, 0);
		for(DirectoryEntry e: root.directories) {
			traverseWrite(e, filenameMap, out);
		}
		for(FileEntry e: root.files) {
			Util.write32(out, filenameMap.get(e.filename));
			Util.write32(out, e.csize);
			Util.write16(out, e.id);
			out.writeByte(0);
			out.writeByte(1);
			Util.write32(out, 0);
		}
	}
	private static void traverseWriteData(DirectoryEntry root, RandomAccessFile out) throws IOException {
		for(DirectoryEntry e: root.directories) {
			traverseWriteData(e, out);
		}
		for(FileEntry e: root.files) {
			e.offset = (int) out.getFilePointer();
			out.write(e.data);
			while(out.getFilePointer() % 4 > 0) {
				out.writeByte(0);
			}
		}
	}
	private static void traverseWriteOffsets(DirectoryEntry root, RandomAccessFile out) throws IOException {
		for(DirectoryEntry e: root.directories) {
			traverseWriteOffsets(e, out);
		}
		for(FileEntry e: root.files) {
			Util.write32(out, e.offset);
		}
	}
	public static void create(File in, RandomAccessFile out, ProgressDisplay display) throws IOException {
		Util.write32(out, 0x02011234);
		Util.write32(out, 0x00000242);
		display.update("Starting...", true, 0);
		DirectoryEntry root = new DirectoryEntry(in.getName().toUpperCase());
		root.directoryCountTotal++;
		display.update("Compressing file data...", true, 0);
		traverse(in, root);
		Map<String,Integer> filenameMap = writeFilenames(root, out);
		Util.write32(out, root.fileCountTotal);
		display.update("Writing file entries...", true, 0);
		traverseFiles1(root, out, (short) 0);
		Util.write32(out, root.fileCountTotal);
		Util.write32(out, root.directoryCountTotal);
		Util.write32(out, root.fileCountTotal);
		long offsetsStart = out.getFilePointer();
		for(int i = 0; i < root.fileCountTotal; i++) out.writeInt(0);
		display.update("Writing file names...", true, 0);
		traverseWrite(root, filenameMap, out);
		display.update("Writing file data...", true, 0);
		traverseWriteData(root, out);
		out.seek(offsetsStart);
		display.update("Writing data offsets...", true, 0);
		traverseWriteOffsets(root, out);
		display.setMaximumProgress(1);
		display.update("Finished.", true, 1);
	}
}
